CREATE OR REPLACE FUNCTION ogr_fdw_version()
	RETURNS text
	AS 'MODULE_PATHNAME', 'ogr_fdw_version'
	LANGUAGE 'c'
	IMMUTABLE STRICT
	PARALLEL SAFE;

CREATE OR REPLACE FUNCTION ogr_fdw_drivers()
	RETURNS text[]
	AS 'MODULE_PATHNAME', 'ogr_fdw_drivers'
	LANGUAGE 'c'
	IMMUTABLE STRICT
	PARALLEL SAFE;
